/*
 * pWobble.java
 *
 * Copyright(c) 2001, Particle
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

import java.applet.Applet;

import java.awt.Event;
import java.awt.Color;
import java.awt.Image;
import java.awt.Graphics;

import java.io.InputStream;
import java.io.IOException;
import java.io.BufferedInputStream;
import java.io.DataInputStream;

import java.net.URL;

/**
 * simple applet to display a quake1 map
 * Note: intentionally made to be JDK 1.0 compatible.
 *
 * Not designed for speed. Horribly slow. Things to improve:
 *
 * Enable bounding box clipping in the BSP tree (clip them to
 * viewing volume), this will cut out some unseen subtrees of
 * the bsp (less stuff for renderer to care about).
 * (this will very likely increase the download file size; as it
 * needs the bounding box info per bsp node)
 *
 * Enable use of visdata. In addition to the above, this will
 * reduce the number of possibly visible faces even further. This
 * will require addition of visdata to download file, and addition
 * of leaf nodes (this addon likely to double download file size)
 *
 * Using any (or both) of the above, you can avoid having to
 * transform all of the vertices. (right now the matrix transforms
 * all of them, even if they're never actually used in rendering)
 *
 * If that is not enough, improve clipping code.
 *
 * A quick and dirty speedup trick is to reduce the sleep time
 * of the applet, and to not yield the main thread. This will tie
 * up the machine (make it seem really really slow), but the
 * applet will run faster... [I don't like greedy applets, so
 * right now it gives plenty of breething space to the system.]
 *
 * some useful info:
 *  - Uses Matrices for Rotation, Translation, Projection
 *  - Uses RIGHT HANDED coordinate system upto Projection
 *    (projection reflects y axis for screen)
 *  - Stores all data as dedicated arrays. (ie: node_planenum,
 *    node_frontchild,node_backchild,node_firstface, and
 *    node_numfaces, would normally be part of a seperate Node
 *    Object, but having them as primitive arrays is simpler.
 *  - Renders things in back to front order.
 *  - Has no bsp tree leafs (you can't tell if something is a
 *    solid, or water, etc.) Also no visdata.
 *  - Wobble is a synonym for Quake :-)
 *
 * @author Alex S.
 * @version 0.0.1
 */
public class pWobble extends Applet implements Runnable {

    // our running thread...
    private Thread m_thread;

    // flag that map is finished downloading...
    private boolean finishedDownloadingMap = false;
    private String currentlyDownloading = "connecting...";

    // planes; 4 entries per plane: [A]x + [B]y + [C]z = [D]
    private float[] planes;

    // points; 3 entries per point: X Y Z
    // (note: must be homogenized by matrix transforms)
    private int[] vertices;

    // vertices after a matrix transform
    private int[] vertices_transformed;

    // plane that this node used to divide the world by
    private short[] node_planenum;

    // the front children of nodes
    private short[] node_frontchild;

    // the back children of nodes
    private short[] node_backchild;

    // first face for this node (we draw these)
    private short[] node_firstface;

    // number of faces for this node (we need to know when to stop)
    private byte[] node_numfaces;

    // what side of the plane is this face on? (does back face culling)
    private byte[] face_side;

    // first edge of this surface edge of this face
    private short[] face_firstedge;

    // number of surface edges
    private byte[] face_numedges;

    // front to back; back to front edges
    private short[] edges;

    // surface edges (faces point to these, and these point to edges)
    private short[] surfedges;

    // default map if not specified as applet param
    private String map = "q1map.pbsp";

    // size of applet window
    private int m_width,m_height;

    // mouse x,y
    private int m_mousex,m_mousey;

    // support for double buffer
    private Image m_image;
    private Graphics m_graphics;

    // player coordinates
    private float player_x,player_y,player_z;

    // player orientation
    private float player_yaw,player_pitch,player_roll;

    // transform matrix
    private float[] tmatrix;

    // projection matrix
    private float[] pmatrix;

    // global matrix (to avoid accessing arrays)
    private float gm0,gm1,gm2,gm3,gm4,gm5,gm6,gm7,gm8,gm9,gm10,gm11,gm12,gm13,gm14,gm15;

    // used for java's fillpoly thingie... (and some quick & dirty clipping)
    private int[] x;
    private int[] y;
    private int[] z;

    // clippping ...
    private int[] cx;
    private int[] cy;
    private int[] cz;

    // general purpose keys; true if pressed...
    private boolean K_FORWARD,K_BACKARD,K_MOVELEFT,K_MOVERIGHT,
    K_ROLLLEFT,K_ROLLRIGHT,K_TURNLEFT,K_TURNRIGHT;

    // object to provide synchronization
    private Object synchobject = new Object();

    // so that we won't have to cast it to float every time
    private static final float PI = (float)Math.PI;

    // screen rectangle
    int minx,maxx,miny,maxy;

    /**
     * initialize the applet
     */
    public void init(){
        String param = getParameter("map");
        if (param != null)
            map = param;

        // need to be JDK 1.0 compatible
        m_width = size().width;
        m_height = size().height;

        // screen rectangle
        minx = 10;
        maxx = m_width-10;
        miny = 45;
        maxy = m_height-10;

        // create double buffer image
        m_image = createImage(m_width,m_height);

        // get graphics object to our double buffer
        m_graphics = m_image.getGraphics();

        synchronized(synchobject){
            // initial player coords (good for default map)
            player_x = 140;
            player_y = 825;
            player_z = -170;
            player_yaw = PI/2;
            player_pitch = 0;
            player_roll = 0;
        }

        // init some arrays
        tmatrix = new float[16];
        pmatrix = new float[16];
        x = new int[128];
        y = new int[128];
        z = new int[128];
        cx = new int[128];
        cy = new int[128];
        cz = new int[128];
    }

    /**
     * thread start method
     */
    public void start(){
        if (m_thread == m_thread) {
            m_thread = new Thread(this);
            m_thread.start();
        }
    }

    /**
     * thread runner (Runnable)
     */
    public void run(){

        // first download the map using this thread...
        try {
            loadAll(new URL(getDocumentBase(),map).openStream());
        } catch (Exception e) {
            e.printStackTrace(); // amm... kinda ignore it.
        }
        vertices_transformed = new int[vertices.length];
        finishedDownloadingMap = true;

        // then loop displaying the map and moving around...
        int step = 20;
        for (;;) {
            repaint();
            Thread.yield(); // let other apps have some fun
            synchronized(synchobject){
                if (K_MOVELEFT) {
                    double tmp = player_yaw + PI/2;
                    player_x += (float)Math.cos(tmp) * step;
                    player_y += (float)Math.sin(tmp) * step;
                }
                if (K_MOVERIGHT) {
                    double tmp = player_yaw - PI/2;
                    player_x += (float)Math.cos(tmp) * step;
                    player_y += (float)Math.sin(tmp) * step;
                }
                if (K_FORWARD) {
                    float sy = (float)Math.sin(player_yaw);
                    float cy = (float)Math.cos(player_yaw);
                    float sp = (float)Math.sin(player_pitch);
                    float cp = (float)Math.cos(player_pitch);
                    player_x += cp*cy * step;
                    player_y += cp*sy * step;
                    player_z += sp * step;
                }
                if (K_BACKARD) {
                    float sy = (float)Math.sin(player_yaw);
                    float cy = (float)Math.cos(player_yaw);
                    float sp = (float)Math.sin(player_pitch);
                    float cp = (float)Math.cos(player_pitch);
                    player_x -= cp*cy * step;
                    player_y -= cp*sy * step;
                    player_z -= sp * step;
                }
                if (K_ROLLLEFT) {
                    player_roll += PI/256;
                    if (player_roll > PI*2)
                        player_roll -= PI*2;
                }
                if (K_ROLLRIGHT) {
                    player_roll -= PI/256;
                    if (player_roll < 0)
                        player_roll += PI*2;
                }
                if (K_TURNLEFT) {
                    player_yaw += PI/64;
                    if (player_yaw > PI*2)
                        player_yaw -= PI*2;
                }
                if (K_TURNRIGHT) {
                    player_yaw -= PI/64;
                    if (player_roll < 0)
                        player_roll += PI*2;
                }
            }
            Thread.yield(); // don't take over the whole machine
            try {
                Thread.sleep(30);   // give it some reasonable time off
            } catch (java.lang.InterruptedException e) {
//                e.printStackTrace();
            }
        }
    }

    /**
     * crude clipping... x,y points to screen.
     * (very inefficient)
     */
    private int clipToScreen(int n){

        float d;
        int i,next,prev,dist,c;
        int[] tmp;
        int abs;

        // minx clip
        c = 0;
        for (i=0;i<n;i++) {
            if (x[i] >= minx) { // point is on right side
                cx[c] = x[i];
                cy[c++] = y[i];
                continue;
            }
            next = (i+1) % n;
            prev = (i+n-1) % n;
            dist = x[i] - minx;
            if (dist < 0)
                dist = -dist;
            if (x[prev] >= minx) {
                abs = x[prev]-minx;
                if (abs < 0)
                    abs = -abs;
                d = (float)dist / (float)(dist + abs);
                cx[c] = minx;
                cy[c++] = y[i]+(int)(d*(y[prev]-y[i]));
            }
            if (x[next] >= minx) {
                abs = x[next]-minx;
                if (abs < 0)
                    abs = -abs;
                d = (float)dist / (float)(dist + abs);
                cx[c] = minx;
                cy[c++] = y[i]+(int)(d*(y[next]-y[i]));
            }
        }
        if (c == 0)
            return c;
        n = c;

        tmp = x;
        x = cx;
        cx = tmp;
        tmp = y;
        y = cy;
        cy = tmp;

        // miny clip
        c = 0;
        for (i=0;i<n;i++) {
            if (y[i] >= miny) { // point is on right side
                cx[c] = x[i];
                cy[c++] = y[i];
                continue;
            }
            next = (i+1) % n;
            prev = (i+n-1) % n;
            dist = y[i] - miny;
            if (dist < 0)
                dist = -dist;
            if (y[prev] >= miny) {
                abs = y[prev]-miny;
                if (abs < 0)
                    abs = -abs;
                d = (float)dist / (float)(dist + abs);
                cx[c] = x[i]+(int)(d*(x[prev]-x[i]));
                cy[c++] = miny;
            }
            if (y[next] >= miny) {
                abs = y[next]-miny;
                if (abs < 0)
                    abs = -abs;
                d = (float)dist / (float)(dist + abs);
                cx[c] = x[i]+(int)(d*(x[next]-x[i]));
                cy[c++] = miny;
            }
        }
        if (c == 0)
            return c;
        n = c;

        tmp = x;
        x = cx;
        cx = tmp;
        tmp = y;
        y = cy;
        cy = tmp;

        // maxx clip
        c = 0;
        for (i=0;i<n;i++) {
            if (x[i] < maxx) { // point is on right side
                cx[c] = x[i];
                cy[c++] = y[i];
                continue;
            }
            next = (i+1) % n;
            prev = (i+n-1) % n;
            dist = x[i] - maxx;
            if (dist < 0)
                dist = -dist;
            if (x[prev] < maxx) {
                abs = x[prev]-maxx;
                if (abs < 0)
                    abs = -abs;
                d = (float)dist / (float)(dist + abs);
                cx[c] = maxx;
                cy[c++] = y[i]+(int)(d*(y[prev]-y[i]));
            }
            if (x[next] < maxx) {
                abs = x[next]-maxx;
                if (abs < 0)
                    abs = -abs;
                d = (float)dist / (float)(dist + abs);
                cx[c] = maxx;
                cy[c++] = y[i]+(int)(d*(y[next]-y[i]));
            }
        }
        if (c == 0)
            return c;
        n = c;

        tmp = x;
        x = cx;
        cx = tmp;
        tmp = y;
        y = cy;
        cy = tmp;

        // maxy clip
        c = 0;
        for (i=0;i<n;i++) {
            if (y[i] < maxy) { // point is on right side
                cx[c] = x[i];
                cy[c++] = y[i];
                continue;
            }
            next = (i+1) % n;
            prev = (i+n-1) % n;
            dist = y[i] - maxy;
            if (dist < 0)
                dist = -dist;
            if (y[prev] < maxy) {
                abs = y[prev]-maxy;
                if (abs < 0)
                    abs = -abs;
                d = (float)dist / (float)(dist + abs);
                cx[c] = x[i]+(int)(d*(x[prev]-x[i]));
                cy[c++] = maxy;
            }
            if (y[next] < maxy) {
                abs = y[next]-maxy;
                if (abs < 0)
                    abs = -abs;
                d = (float)dist / (float)(dist + abs);
                cx[c] = x[i]+(int)(d*(x[next]-x[i]));
                cy[c++] = maxy;
            }
        }
        if (c == 0)
            return c;

        tmp = x;
        x = cx;
        cx = tmp;
        tmp = y;
        y = cy;
        cy = tmp;

        return c;
    }


    /**
     * very very crude clipping...
     * clips the x,y,z points to the z-axis (only negative z stays)
     */
    private int clipToZ(int n){
        // clip to Z
        float d;
        int i,next,prev, c = 0;  // good point count
        for (i=0;i<n;i++) {
            if (z[i] <= 0) {  // point is on right side
                cx[c] = x[i];
                cy[c] = y[i];
                cz[c] = z[i];
                c++;
                continue;
            }
            next = (i+1) % n;
            prev = (i+n-1) % n;
            if (z[prev] <= 0) {
                // this is first point outside, create a split point
                d = (float)z[i] / (float)(z[i] - z[prev]);  // d is positive.
                cx[c] = x[i]+(int)(d*(x[prev]-x[i]));
                cy[c] = y[i]+(int)(d*(y[prev]-y[i]));
                cz[c] = 0;
                c++;
            }
            if (z[next] <= 0) {
                // next point is ok, we'll handle it now
                // z[i] is positive, z[next] is negative (or 0).
                d = (float)z[i] / (float)(z[i] - z[next]);  // d is positive.
                cx[c] = x[i]+(int)(d*(x[next]-x[i]));
                cy[c] = y[i]+(int)(d*(y[next]-y[i]));
                cz[c] = 0;
                c++;
            }
        }
        int[] tmp;
        tmp = x;
        x = cx;
        cx = tmp;
        tmp = y;
        y = cy;
        cy = tmp;
        tmp = z;
        z = cz;
        cz = tmp;
        return c;
    }

    /**
     * renders a particular face
     *
     * @param face Face number to render
     */
    private void paintFace(int face) {
        int n = face_numedges[face];
        int firstedge = face_firstedge[face];
        int i,p,edge,clipz;
        boolean allnegative = true,allpositive = true;
        for (i=0; i < n; i++) {
            edge = surfedges[firstedge++];
            p = edge < 0 ? edges[((-edge) << 1)|1]:edges[edge << 1];
            x[i] = (int)vertices_transformed[p*=3];
            y[i] = (int)vertices_transformed[p+1];
            z[i] = clipz = (int)vertices_transformed[p+2];

            // clipping
            // trivial cases, all points have positive or all negative z
            if (clipz > 0)
                allnegative = false;
            else
                allpositive = false;
        }
        // trivial rejection
        if (allpositive)
            return;
        if (!allnegative)
            n = clipToZ(n);
        if (n == 0)
            return;

        // project face coordinates
        matrixTransformPerspective(x,y,z,n);

        n = clipToScreen(n);
        if (n == 0)
            return;
        // render
        m_graphics.fillPolygon(x,y,n);

    }

    /**
     * loops through node's faces; and renders them if
     * they're facing us (does back face culling)
     *
     * @param node Node number where to grab faces
     * @param side Which side of the face are we on?
     */
    private void paintNodeFaces(int node,int side){
        int color = node_planenum[node];
        // set an acceptable [greenish] color for these faces
        m_graphics.setColor(new Color((byte)(color%12),
                                      (byte)(color%126),
                                      (byte)(color%12)));
        int face = node_firstface[node];
        int n = node_numfaces[node] + face;
        for (;face<n;face++)
            if (face_side[face] == side)    // back face culling
                paintFace(face);
    }

    /**
     * traverse the bsp tree, painting it in back to front order
     *
     * @param node The number of current node.
     */
    private void paintBSP(int node){
        if (node >= 0) {
            int p = node_planenum[node] * 4;
            if (planes[p]*player_x + planes[p+1]*player_y + planes[p+2]*player_z < planes[p+3]) {
                // if back; draw front, then current, then back
                paintBSP(node_frontchild[node]);
                paintNodeFaces(node,1); // side is back
                paintBSP(node_backchild[node]);
            } else {
                // if front, draw back, then current, then front
                paintBSP(node_backchild[node]);
                paintNodeFaces(node,0);   // side is front
                paintBSP(node_frontchild[node]);
            }
        }
    }

    /**
     * the paint method...
     *
     * @param g The Graphics context into which to draw.
     */
    public void paint(Graphics g){

        // clear double buffer
        m_graphics.setColor(Color.darkGray);
        m_graphics.fillRect(0,0,m_width,m_height);

        if (finishedDownloadingMap) {
            // world to view transform

            matrixUnit(tmatrix);

            synchronized(synchobject){
                // player rotations
                matrixRotateZ(tmatrix,player_roll);
                matrixRotateX(tmatrix,-player_pitch);
                matrixRotateY(tmatrix,-player_yaw);

                // correct for coordinate system differences
                matrixRotateX(tmatrix,-PI/2);
                matrixRotateZ(tmatrix,PI/2);

                // player location...
                matrixTranslate(tmatrix,-player_x,-player_y,-player_z);
            }

            // created rotation matrix... transform all vertices
            matrixTransformBasic(tmatrix,vertices,
                                 vertices_transformed,
                                 vertices.length);

            // screen/perspective

            matrixUnit(pmatrix);

            // move to center of screen
            matrixTranslate(pmatrix,m_width/2,m_height/2,0);

            // 90 degrees field of view
            matrixPerspective(pmatrix,m_width/2,1);

            // on screen, Y increases downward
            matrixReflectY(pmatrix);

            // store this matrix as global variables
            matrixToGlobal(pmatrix);

            // paint bsp, starting at root node (node 0)
            paintBSP(0);

            // write some debug info...
            synchronized(synchobject){
                m_graphics.setColor(Color.white);
                float tmpx = ((int)(player_x * 10))/10.0F;
                float tmpy = ((int)(player_y * 10))/10.0F;
                float tmpz = ((int)(player_z * 10))/10.0F;
                m_graphics.drawString("("+tmpx+", "+tmpy+", "+tmpz+")",10,10);
                float deg = ((int)((player_yaw * 180/PI)*10))/10.0F;
                m_graphics.drawString("yaw: "+deg,10,20);
                deg = ((int)((player_pitch * 180/PI)*10))/10.0F;
                m_graphics.drawString("pitch: "+deg,10,30);
                deg = ((int)((player_roll * 180/PI)*10))/10.0F;
                m_graphics.drawString("roll: "+deg,10,40);
            }

        } else {
            // haven't downloaded the map yet... display some donwload info
            m_graphics.setColor(Color.white);
            m_graphics.drawString("Downloading map: "+currentlyDownloading,10,10);
        }

        // draw a cool black rectangle :-)
        m_graphics.setColor(Color.black);
        m_graphics.drawRect(minx,miny,maxx-minx,maxy-miny);

        // blit our double buffer image to screen
        g.drawImage(m_image,0,0,null);
    }

    /**
     * the update, called every time "repaint()" is called,
     * or when ImageObserver stuff happens...
     *
     * recieves a graphics context, with clipping regions
     * set to the "update" region...
     *
     * @param g The graphics context into which to draw.
     */
    public void update(Graphics g){
        // paint map...
        paint(g);
    }


    /**
     * a key handler...
     * [JDK 1.0 Event Handling]
     */
    public boolean keyUp(Event evt,int key){
        if (key == Event.UP || key == 'w' || key == 'W') {
            K_FORWARD = false;
        } else if (key == Event.DOWN || key == 's' || key == 'S') {
            K_BACKARD = false;
        } else if (key == Event.LEFT) {
            K_TURNLEFT = false;
        } else if (key == 'a' || key == 'A') {
            K_MOVELEFT = false;
        } else if (key == Event.RIGHT) {
            K_TURNRIGHT = false;
        } else if (key == 'd' || key == 'D') {
            K_MOVERIGHT = false;
        } else if (key == 'q' || key == 'Q') {
            K_ROLLLEFT = false;
        } else if (key == 'e' || key == 'E') {
            K_ROLLRIGHT = false;
        }
        return true;
    }

    /**
     * a key handler...
     * [JDK 1.0 Event Handling]
     */
    public boolean keyDown(Event evt,int key){
        if (key == Event.UP || key == 'w' || key == 'W') {
            K_FORWARD = true;
        } else if (key == Event.DOWN || key == 's' || key == 'S') {
            K_BACKARD = true;
        } else if (key == Event.LEFT) {
            K_TURNLEFT = true;
        } else if (key == 'a' || key == 'A') {
            K_MOVELEFT = true;
        } else if (key == Event.RIGHT) {
            K_TURNRIGHT = true;
        } else if (key == 'd' || key == 'D') {
            K_MOVERIGHT = true;
        } else if (key == 'q' || key == 'Q') {
            K_ROLLLEFT = true;
        } else if (key == 'e' || key == 'E') {
            K_ROLLRIGHT = true;
        }
        return true;
    }

    /**
     * capture even when mouse is pressed...
     * [JDK 1.0 Event Handling]
     *
     * @param evt The event...
     * @param x the x mouse position.
     * @param y the y mouse posiiton.
     * @return whether or not we handled the event.
     */
    public boolean mouseDown(Event evt, int x, int y){
        m_mousex = x;
        m_mousey = y;
        return true;
    }

    /**
     * capture mouse release..
     * [JDK 1.0 Event Handling]
     *
     * @param evt The event...
     * @param x the x mouse position.
     * @param y the y mouse posiiton.
     * @return whether or not we handled the event.
     */
    public boolean mouseUp(Event evt, int x, int y){
        m_mousex = x;
        m_mousey = y;
        return true;
    }

    /**
     * capture mouse drag...
     * [JDK 1.0 Event Handling]
     *
     * @param evt The event...
     * @param x the x mouse position.
     * @param y the y mouse posiiton.
     * @return whether or not we handled the event.
     */
    public boolean mouseDrag(Event evt, int x, int y){
        synchronized(synchobject){
            if (m_mousex < x) {
                /* left to right movement */
                player_yaw -= (x - m_mousex) * PI/256;
                if (player_yaw < 0)
                    player_yaw += PI*2;
            } else if (m_mousex > x) {
                /* right to left movement */
                player_yaw += (m_mousex - x) * PI/256;
                if (player_yaw > PI*2)
                    player_yaw -= PI*2;
            }
            /* up to down movement */
            if (m_mousey < y) {
                player_pitch -= (y - m_mousey) * PI/256;
                if (player_pitch < -PI/2)
                    player_pitch = -PI/2;
                /* down to up movement */
            } else if (m_mousey > y) {
                player_pitch += (m_mousey - y) * PI/256;
                if (player_pitch > PI/2)
                    player_pitch = PI/2;
            }
        }
        m_mousex = x;
        m_mousey = y;
        return true;
    }


    /**
     * turns this matrix into global matrix (to avoid
     * accessing matrix arrays everytime)
     */
    private void matrixToGlobal(float[] m){
        gm0 = m[0]; gm1 = m[1]; gm2 = m[2]; gm3 = m[3];
        gm4 = m[4]; gm5 = m[5]; gm6 = m[6]; gm7 = m[7];
        gm8 = m[8]; gm9 = m[9]; gm10 = m[10]; gm11 = m[11];
        gm12 = m[12]; gm13 = m[13]; gm14 = m[14]; gm15 = m[15];
    }

    /**
     * 4x4 matrix: make unit
     */
    private static void matrixUnit(float[] m){
        for (int i=0;i<4;i++)
            for (int j=0;j<4;j++)
                m[(i<<2) + j] = i == j ? 1:0;
    }

    /**
     * 4x4 matrix translation
     */
    private static void matrixTranslate(float[] m,float x,float y,float z){
        for (int i=0;i<4;i++)
            m[(i<<2)+3] += m[i<<2]*x + m[(i<<2)+1]*y + m[(i<<2)+2]*z;
    }

    /**
     * 4x4 matrix rotation; theta is in radiants.
     *
     * newx = x
     * newy = cos * y - sin * z
     * newz = sin * y + cos * z
     */
    private static void matrixRotateX(float[] m,float theta){
        float t1,t2;
        float cos = (float)Math.cos(theta);
        float sin = (float)Math.sin(theta);
        for (int i=0;i<4;i++) {
            t1 = m[(i<<2)+1];
            t2 = m[(i<<2)+2];
            m[(i<<2)+1] = t1*cos+t2*sin;
            m[(i<<2)+2] = t1*-sin+t2*cos;
        }
    }

    /**
     * 4x4 matrix rotation; theta is in radiants.
     *
     * newx = cos * x + sin * z
     * newy = y
     * newz = -sin * x + cos * z
     */
    private static void matrixRotateY(float[] m,float theta){
        float t0,t2;
        float cos = (float)Math.cos(theta);
        float sin = (float)Math.sin(theta);
        for (int i=0;i<4;i++) {
            t0 = m[(i<<2)];
            t2 = m[(i<<2)+2];
            m[(i<<2)] = t0*cos + t2*-sin;
            m[(i<<2)+2] = t0*sin + t2*cos;
        }
    }

    /**
     * 4x4 matrix rotation; theta is in radiants.
     *
     * newx = cos * x - sin * y
     * newy = sin * x + cos * y
     * newz = z
     */
    private static void matrixRotateZ(float[] m,float theta){
        float t0,t1;
        float cos = (float)Math.cos(theta);
        float sin = (float)Math.sin(theta);
        for (int i=0;i<4;i++) {
            t0 = m[(i<<2)];
            t1 = m[(i<<2)+1];
            m[(i<<2)] = t0*cos+t1*sin;
            m[(i<<2)+1] = t0*-sin+t1*cos;
        }
    }

    /**
     * 4x4 matrix reflection in XZ plane
     */
    private static void matrixReflectY(float[] m){
        for (int i=0;i<4;i++)
            m[(i<<2)+1] = -m[(i<<2)+1];
    }

    /**
     * create perspective matrix
     *
     * @param zrpr at what z is the eye?
     * @param zvp at what z is the view plane?
     */
    private static void matrixPerspective(float[] m,float zprp,float zvp){
        float t2,t3;
        float d1 = zprp - zvp;
        float a = -zvp/d1;
        float b = -1/d1;
        float c = zvp*(zprp/d1);
        float d = zvp/d1;
        for (int i=0;i<4;i++) {
            t2 = m[(i<<2)+2];
            t3 = m[(i<<2)+3];
            m[(i<<2)+2] = t2*a + t3*b;
            m[(i<<2)+3] = t2*c + t3*d;
        }
    }

    /**
     * transforms an array of vertices using the matrix
     * only outputs x and y, z remains the same...
     * uses the global matrix...
     * (unrolled for speed)
     */
    private void matrixTransformPerspective(int[] a,int[] b,int[] c,int n){
        float ax,ay,az,w;
        for (int i=0;i<n;i++) {
            ax = a[i];
            ay = b[i];
            az = c[i];
            w = gm12 * ax + gm13 * ay + gm14 * az + gm15;
            a[i] = (int)((gm0 * ax + gm1 * ay + gm2 * az + gm3) / w);
            b[i] = (int)((gm4  * ax + gm5  * ay + gm6  * az + gm7) / w);
            c[i] = (int)((gm8  * ax + gm9  * ay + gm10 * az + gm11)/ w);
        }
    }


    /**
     * transforms an array of vertices using the matrix
     * (unrolled for speed)
     * [doesn't do a *correct* transform; must use a more correct
     * transform for perspective...]
     */
    private static void matrixTransformBasic(float[] m,int[] a,int[] b,int n){
        float ax,ay,az,w;
        float m0 = m[0], m1 = m[1], m2 = m[2], m3 = m[3], m4 = m[4], m5 = m[5],
              m6 = m[6], m7 = m[7], m8 = m[8], m9 = m[9], m10 = m[10], m11 = m[11],
                                                                                                                                         m12 = m[12], m13 = m[13], m14 = m[14], m15 = m[15];
        for (int i=0;i<n;i+=3) {
            ax = a[i];
            ay = a[i+1];
            az = a[i+2];
            b[i] = (int)(m0 * ax + m1 * ay + m2 * az + m3);
            b[i+1] = (int)(m4  * ax + m5  * ay + m6  * az + m7);
            b[i+2] = (int)(m8  * ax + m9  * ay + m10 * az + m11);
        }
    }


    /**
     * reads all data from an input stream
     * [repaint the screen with download info approximately 4 times a second]
     *
     * @param is InputStream from which to read.
     */
    private void loadAll(InputStream is) throws IOException {
        long timer; // repaint timer...
        DataInputStream dis = new DataInputStream(
            new BufferedInputStream(is,8192));  // 8k buffer
        int i,j = dis.readShort();  // numplanes

        currentlyDownloading = " "+j+" planes";
        repaint();
        timer = System.currentTimeMillis() + 250;

        planes = new float[j*=4];
        for (i=0;i<j;i+=4) {
            planes[i]   = (float)((int)dis.readByte()/127.0);
            planes[i+1] = (float)((int)dis.readByte()/127.0);
            planes[i+2] = (float)((int)dis.readByte()/127.0);
            planes[i+3] = (float)dis.readShort();
            if (timer < System.currentTimeMillis()) {
                currentlyDownloading = " "+(i/4)+"/"+(j/4)+" planes";
                repaint();
                timer = System.currentTimeMillis() + 250;
            }
        }
        j = dis.readShort();    // numvertices

        currentlyDownloading = " "+j+" vertices";
        repaint();
        timer = System.currentTimeMillis() + 250;


        vertices = new int[j*=3];
        for (i=0;i<j;i+=3) {
            vertices[i]   = dis.readShort();
            vertices[i+1] = dis.readShort();
            vertices[i+2] = dis.readShort();
            if ((i%2) == 0 && timer < System.currentTimeMillis()) {
                currentlyDownloading = " "+(i/3)+"/"+(j/3)+" vertices";
                repaint();
                timer = System.currentTimeMillis() + 250;
            }
        }
        j = dis.readShort();    // numnodes
        currentlyDownloading = " "+j+" bsp nodes";
        repaint();
        timer = System.currentTimeMillis() + 250;

        node_planenum = new short[j];
        node_frontchild = new short[j];
        node_backchild = new short[j];
        node_firstface = new short[j];
        node_numfaces = new byte[j];
        for (i=0;i<j;i++) {
            node_planenum[i] = dis.readShort();
            node_frontchild[i] = dis.readShort();
            node_backchild[i] = dis.readShort();
            node_firstface[i] = dis.readShort();
            node_numfaces[i] = dis.readByte();

            if ((i%2) == 0 && timer < System.currentTimeMillis()) {
                currentlyDownloading = " "+i+"/"+j+" bsp nodes";
                repaint();
                timer = System.currentTimeMillis() + 250;
            }

        }
        j = dis.readShort();    // numfaces
        currentlyDownloading = " "+j+" faces";
        repaint();
        timer = System.currentTimeMillis() + 250;
        face_side = new byte[j];
        face_firstedge = new short[j];
        face_numedges = new byte[j];
        for (i=0;i<j;i++) {
            face_firstedge[i] = dis.readShort();
            face_numedges[i] = dis.readByte();
            face_side[i] = (byte)((face_numedges[i] & 128) >> 7);
            face_numedges[i] &= 127;
            if ((i%2) == 0 && timer < System.currentTimeMillis()) {
                currentlyDownloading = " "+i+"/"+j+" faces";
                repaint();
                timer = System.currentTimeMillis() + 250;
            }
        }
        j = dis.readShort();    // numedges
        currentlyDownloading = " "+j+" edges";
        repaint();
        timer = System.currentTimeMillis() + 250;
        edges = new short[j*=2];
        for (i=0;i<j;i+=2) {
            edges[i] = dis.readShort();
            edges[i+1] = dis.readShort();
            if ((i%2) == 0 && timer < System.currentTimeMillis()) {
                currentlyDownloading = " "+(i/2)+"/"+(j/2)+" edges";
                repaint();
                timer = System.currentTimeMillis() + 250;
            }
        }
        j = dis.readShort();    // surface edges
        currentlyDownloading = " "+j+" surface edges";
        repaint();
        timer = System.currentTimeMillis() + 250;
        surfedges = new short[j];
        for (i=0;i<j;i++) {
            surfedges[i] = dis.readShort();
            if ((i%2) == 0 && timer < System.currentTimeMillis()) {
                currentlyDownloading = " "+i+"/"+j+" surface edges";
                repaint();
                timer = System.currentTimeMillis() + 250;
            }
        }
    }
}

